__package__ = 'sismic'
__version__ = '1.6.8'
__licence__ = 'LGPL3'
__author__ = 'Alexandre Decan'
__url__ = 'https://github.com/AlexandreDecan/sismic/'
__description__ = 'Sismic Interactive Statechart Model Interpreter and Checker'
